library capsule;

import 'dart:async';
import 'dart:convert';

import 'package:flutter/foundation.dart';
import 'package:flutter_inappwebview/flutter_inappwebview.dart';
import 'package:passkeys_capsule/authenticator.dart';
import 'package:passkeys_capsule/types.dart';
import 'package:uuid/rng.dart';
import 'package:webauthn/webauthn.dart';
import 'dart:io';

import 'create_wallet_result.dart';
import 'environment.dart';
import 'signature_results.dart';
import 'wallet.dart';

part 'integrations/metamask/capsule_signer.dart';
part 'integrations/metamask/typed_data_domain.dart';
part 'integrations/metamask/typed_data_field.dart';

const es256Algorithm = -7;
const rs256Algorithm = -257;

/// Flutter implementation of Capsule API
class Capsule {
  /// Constructs a `Capsule`.
  ///
  /// After constructing, call [init] to initialize the SDK, e.g.
  /// ```
  /// final capsule = Capsule(environment: Environment.beta)..init();
  /// ```
  ///
  /// API key is optional when using `Environment.dev`, but
  /// it required for other environments like beta and prod.
  ///
  /// The jsBridgeUri is optional and can be used to override the default
  /// URI for any of the dev environments. This can be useful when making
  /// changes to the js-bridge and testing them in a local environment, but
  /// is not restricted to that use case.
  ///
  /// Relying party ID is only required for dev environment. Typically, you
  /// can host your own .well-kown files and use ngrok to expose your local
  /// server to the internet. This will be necessary to test with passkeys
  /// locally using a dev environment.
  Capsule({
    required this.environment,
    required this.apiKey,
    this.jsBridgeUri,
    this.relyingPartyId,
  });

  /// The environment to use to configure the Capsule SDK.
  final Environment environment;

  /// API key to use to configure the Capsule SDK.
  ///
  /// If you don't have one, email hello@usecapsule.com.
  final String apiKey;

  final Uri? jsBridgeUri;
  final String? relyingPartyId;

  HeadlessInAppWebView? _headlessInAppWebView;
  InAppWebViewController? _webViewController;
  final Completer<void> _initCompleter = Completer();
  final ChromeSafariBrowser _browser = ChromeSafariBrowser();
  final StreamController<List<dynamic>> _asyncResultsController =
      StreamController.broadcast();
  late final Stream<List<dynamic>> _asyncResults =
      _asyncResultsController.stream;

  Uri _jsBridgeUri() {
    if (jsBridgeUri != null) {
      return jsBridgeUri!;
    }

    switch (environment) {
      case Environment.dev:
        return Uri.parse('http://localhost:3004/');
      case Environment.sandbox:
        return Uri.parse('https://js-bridge.sandbox.usecapsule.com/');
      case Environment.beta:
        return Uri.parse('https://js-bridge.beta.usecapsule.com/');
      case Environment.prod:
        return Uri.parse('https://js-bridge.prod.usecapsule.com/');
    }
  }

  String _relyingPartyId() {
    if (relyingPartyId != null) {
      return relyingPartyId!;
    }

    switch (environment) {
      case Environment.dev:
        throw Exception(
            'Custom relying party ID is required for dev environment');
      case Environment.sandbox:
        return 'app.sandbox.usecapsule.com';
      case Environment.beta:
        return 'app.beta.usecapsule.com';
      case Environment.prod:
        return 'app.usecapsule.com';
    }
  }

  /// Call this method immediately after constructing an instance.
  ///
  /// It is intentional that no `Future` is returned, because the Flutter
  /// UI should not block on expensive initialization. Calling any async method
  /// of the class will ensure that the initialization has completed.
  void init() {
    if (Platform.isAndroid) {
      AndroidInAppWebViewController.setWebContentsDebuggingEnabled(true);
    }
    _headlessInAppWebView = HeadlessInAppWebView(
      onConsoleMessage: (controller, consoleMessage) {
        debugPrint("CONSOLE MESSAGE: ${consoleMessage.message}");
      },
      initialUrlRequest: URLRequest(url: _jsBridgeUri()),
      onLoadStop: (controller, url) {
        _webViewController = controller;
        controller.addJavaScriptHandler(
          handlerName: 'open',
          callback: (args) {
            _browser.open(url: Uri.parse(args[0].toString()));
          },
        );
        controller.addJavaScriptHandler(
          handlerName: 'asyncResult',
          callback: (args) {
            _asyncResultsController.add(args);
          },
        );
        controller.evaluateJavascript(source: """
          window.postMessage({
            messageType: 'Capsule#init',
            arguments: {
              environment: '${environment.name.toUpperCase()}',
              apiKey: '$apiKey',
            }
          });
        """);
        _initCompleter.complete();
      },
    );
    _headlessInAppWebView!.run();
  }

  Future<dynamic> _postMessage(
    String methodName, [
    List<dynamic>? arguments = const [],
  ]) async {
    await _initCompleter.future;
    final webViewController = _webViewController;
    if (webViewController == null) {
      throw Exception("webViewController not initialized");
    }
    Completer<List<dynamic>> completer = Completer();
    StreamSubscription<List<dynamic>>? subscription =
        _asyncResults.listen((event) {
      completer.complete(event);
    });
    String encodedArguments = jsonEncode(arguments);
    _webViewController?.evaluateJavascript(source: """
      window.postMessage({
        'messageType': 'Capsule#invokeMethod',
        'methodName': '$methodName',
        'arguments': $encodedArguments
      });
    """);
    final result = await completer.future;
    subscription.cancel();
    return result;
  }

  /// Clears storage.
  ///
  /// `keepSecretKey` will retain the Paillier secret key if true.
  Future<void> clearStorage(bool keepSecretKey) async {
    await _postMessage('clearStorage', [keepSecretKey]);
  }

  /// Sets the email of the currently logged in user.
  Future<void> setEmail(String email) async {
    await _postMessage('setEmail', [email]);
  }

  /// Gets the email of the currently logged in user, if one exists.
  Future<String?> getEmail() async {
    final result = await _postMessage('getEmail');
    return result.first;
  }

  /// Gets the wallets of the currently logged in user, if one exists.
  ///
  /// Not implemented yet.
  Future<Map<String, Wallet>> getWallets() async {
    final result = await _postMessage('getWallets');
    return result.first ?? {};
  }

  Future<dynamic> login() async {
    final getWebChallengeResult = await _postMessage('getWebChallenge');

    final challenge = getWebChallengeResult.first["challenge"] as String;

    final passkeyAuthenticator = PasskeyAuthenticator();

    final authenticationRequest = AuthenticateRequestType(
        relyingPartyId: _relyingPartyId(),
        challenge: challenge,
        timeout: 60000,
        userVerification: null,
        allowCredentials: null,
        mediation: MediationType.Required);

    final authenticateRequestResponse =
        await passkeyAuthenticator.authenticate(authenticationRequest);

    final verifyWebChallengeResult = await _postMessage('verifyWebChallenge', [
      authenticateRequestResponse.id,
      authenticateRequestResponse.authenticatorData,
      authenticateRequestResponse.clientDataJSON,
      authenticateRequestResponse.signature,
    ]);

    final userId = verifyWebChallengeResult.first['data']["userId"];

    final wallets = await _postMessage('login', [
      userId,
      authenticateRequestResponse.id,
      authenticateRequestResponse.userHandle,
    ]);

    return wallets;
  }

  Future<void> generatePasskey(String email, String biometricsId) async {
    final userHandle = CryptoRNG().generate();

    final userHandleEncoded = base64Encode(userHandle);

    final passkeyAuthenticator = PasskeyAuthenticator();
    final registerRequest = RegisterRequestType(
      challenge: "LTCT/hWLtJenIgi0oUhkJz7dE8ng+pej+i6YI1QQu60=",
      relyingParty: RelyingPartyType(name: 'Capsule', id: _relyingPartyId()),
      user: UserType(displayName: email, name: email, id: userHandleEncoded),
      authSelectionType: AuthenticatorSelectionType(
          authenticatorAttachment: 'platform',
          requireResidentKey: true,
          residentKey: 'required',
          userVerification: 'required'),
      pubKeyCredParams: [
        PubKeyCredParamType(
            type: PublicKeyCredentialType.publicKey.value, alg: es256Algorithm),
        PubKeyCredParamType(
            type: PublicKeyCredentialType.publicKey.value, alg: rs256Algorithm)
      ],
      timeout: 60000,
      attestation: 'direct',
      excludeCredentials: [],
    );

    final registerRequestResponse =
        await passkeyAuthenticator.register(registerRequest);

    await _postMessage('generatePasskey', [
      registerRequestResponse.attestationObject,
      registerRequestResponse.clientDataJSON,
      registerRequestResponse.id,
      userHandle,
      biometricsId,
    ]);

    return;
  }

  /// Fetches wallets for the currently logged in user.
  ///
  /// Not implemented yet.
  Future<List> fetchWallets() async {
    final result = await _postMessage('fetchWallets');
    return result.first ?? [];
  }

  /// Determines if a user exists with the given email address.
  Future<bool> checkIfUserExists(String email) async {
    final result = await _postMessage('checkIfUserExists', [email]);
    return result.first ?? false;
  }

  /// Creates a new user with the given email address.
  ///
  /// This kicks off a verification process that will
  /// send an email to the user. Call `verifyEmail` with
  /// the verification code.
  Future<void> createUser(String email) async {
    await _postMessage('createUser', [email]);
  }

  /// Supplies email verification code obtained from prompting the user.
  Future<String> verifyEmail(String verificationCode) async {
    final result = await _postMessage('verifyEmail', [verificationCode]);
    final uri = Uri.parse(result[0]);
    return uri.pathSegments.last;
  }

  /// Sets up 2FA for the user.
  Future<String> setup2FA() async {
    final result = await _postMessage('setup2FA');
    return result['uri'];
  }

  /// Turns on 2FA for the user.
  Future<void> enable2FA(String verificationCode) async {
    return await _postMessage('enable2FA', [verificationCode]);
  }

  /// Checks if 2FA is enabled for the user.
  Future<bool> check2FAStatus() async {
    final result = await _postMessage('resendVerificationCode');
    return result['isSetup'];
  }

  /// Sends another verification code.
  Future<void> resendVerificationCode() async {
    return await _postMessage('resendVerificationCode');
  }

  /// Returns true if the user is fully logged in.
  Future<bool> isFullyLoggedIn() async {
    final result = await _postMessage('isFullyLoggedIn');
    return result.first ?? false;
  }

  /// Checks if an authenticated session exists.
  Future<bool> isSessionActive() async {
    final result = await _postMessage('isSessionActive');
    return result.first ?? false;
  }

  /// Distributes a new wallet share, returning the recovery share.
  Future<String> distributeNewWalletShare({
    required String walletId,
    required String userShare,
  }) async {
    return await _postMessage('distributeNewWalletShare', [
      walletId,
      userShare,
    ]) as String;
  }

  /// Creates a `Wallet` for the user.
  ///
  /// `customFunction` is not implemented.
  Future<CreateWalletResult> createWallet({
    required bool skipDistribute,
  }) async {
    final result = await _postMessage('createWallet', [skipDistribute]);
    final wallets = result[0];
    return CreateWalletResult(
      wallet: Wallet.from(wallets[0]['wallet']),
      recoveryShare: wallets[0]['recoveryShare'],
    );
  }

  /// Signs a message.
  Future<FullSignatureResult> signMessage({
    required String walletId,
    required String messageBase64,
  }) async {
    final result = await _postMessage('signMessage', [
      walletId,
      messageBase64,
    ]);
    return FullSignatureResult.fromMap(result[0]);
  }

  /// Signs a transaction.
  Future<FullSignatureResult> signTransaction({
    required String walletId,
    required String rlpEncodedTxBase64,
    required String chainId,
  }) async {
    final result = await _postMessage('signTransaction', [
      walletId,
      rlpEncodedTxBase64,
      chainId,
    ]);
    return FullSignatureResult.fromMap(result[0]);
  }

  /// Sends a transaction.
  Future<FullSignatureResult> sendTransaction({
    required String walletId,
    required String rlpEncodedTxBase64,
    required String chainId,
  }) async {
    final result = await _postMessage('sendTransaction', [
      walletId,
      rlpEncodedTxBase64,
      chainId,
    ]);
    return FullSignatureResult.fromMap(result[0]);
  }

  /// Logs the user out.
  Future<void> logout() => _postMessage('logout');

  /// Instructs the SDK to dispose of any resources
  /// that require explicit disposal.
  void dispose() {
    _webViewController = null;
    _headlessInAppWebView?.dispose();
    _headlessInAppWebView = null;
  }
}
